from typing import Union, List, Any, Optional, TypeVar, Type
from decimal import Decimal
from otree.currency import RealWorldCurrency, Currency
from otree.database import AnyModel

class Currency(Currency):
    """
    PyCharm autocomplete seems to require that I explicitly define the class in this file
    (if I import, it says the reference to Currency is not found)
    """

cu = Currency

class DecimalValue(Decimal):
    _config: str
    def __new__(
        cls, amount: Union[int, float, Decimal], config: str
    ) -> 'DecimalValue': ...
    def _format_for_display(self) -> str: ...

def dec(amount: Union[int, float, Decimal], config: str) -> DecimalValue:
    pass

def currency_range(first, last, increment) -> List[Currency]:
    pass

# mocking the public API for PyCharm autocomplete.
# one downside is that PyCharm doesn't seem to fully autocomplete arguments
# in the .pyi. It gives the yellow pop-up, but doesn't complete what you
# are typing. (2017-07-01: seems to work in PyCharm 2017.1.4?)
class models:
    """
    The code in this class has nothing to do with implementation,
    but rather defines the interface for model fields,
    so that pyCharm autocompletes them properly.

    It defines their __init__ so that when instantiating the class,
    PyCharm suggests the right arguments.
    Apart from that, they can be used as the equivalent Python data type
    (e.g. BooleanField is a bool, CharField is str)

    Without inheriting from bool, str, etc., PyCharm flags certain usages in yellow,
    like:

        c(1) + c(1)

    Results in: "Currency does not define __add__, so the + operator cannot
    be used on its instances"

    If "a" is a CurrencyField, then
        self.a + c(1)

    PyCharm warns: 'Currency does not define __add__, so the + operator cannot
    be used on its instances'

        c(1) + 1

    'Expected type "int", got "Currency" instead'

        self.a + 1

    'Expected type "int", got "CurrencyField" instead'


    """

    def __getattr__(self, item):
        pass

    class BooleanField(bool):
        def __init__(
            self,
            *,
            choices=None,
            widget=None,
            initial=None,
            label=None,
            doc="",
            blank=False,
            **kwargs
        ):
            pass

    class StringField(str):
        def __init__(
            self,
            *,
            choices=None,
            widget=None,
            initial=None,
            label=None,
            doc="",
            max_length=10000,
            blank=False,
            **kwargs
        ):
            pass

    class LongStringField(str):
        def __init__(
            self,
            *,
            initial=None,
            label=None,
            doc="",
            max_length=None,
            blank=False,
            **kwargs
        ):
            pass

    # need to copy-paste the __init__ between
    # Integer, Float, and Currency
    # because if I use inheritance, PyCharm doesn't auto-complete
    # while typing args
    class IntegerField(int):
        def __init__(
            self,
            *,
            choices=None,
            widget=None,
            initial=None,
            label=None,
            doc="",
            min=None,
            max=None,
            blank=False,
            **kwargs
        ):
            pass

    class FloatField(float):
        def __init__(
            self,
            *,
            choices=None,
            widget=None,
            initial=None,
            label=None,
            doc="",
            min=None,
            max=None,
            blank=False,
            **kwargs
        ):
            pass

    class CurrencyField(Currency):
        def __init__(
            self,
            *,
            choices=None,
            widget=None,
            initial=None,
            label=None,
            doc="",
            min=None,
            max=None,
            blank=False,
            **kwargs
        ):
            pass

    class DecimalField(Decimal):
        def __init__(
            self,
            *,
            config_name: str,
            choices=None,
            widget=None,
            initial=None,
            label=None,
            doc="",
            min=None,
            max=None,
            blank=False,
            **kwargs
        ):
            pass

    @staticmethod
    def Link(to) -> Any:
        pass

class widgets:
    def __getattr__(self, item):
        pass

    # don't need HiddenInput because you can just write <input type="hidden" ...>
    # and then you know the element's selector
    class CheckboxInput:
        pass

    class RadioSelect:
        pass

    class RadioSelectHorizontal:
        pass

class BaseConstants:
    pass

class BaseSubsession:
    # mark it as Any so that PyCharm can infer custom fields based on usage instead
    session: Any
    round_number: int
    def get_groups(self) -> List[GroupTV]:
        pass

    def get_group_matrix(self) -> List[List[int]]:
        pass

    def set_group_matrix(self, group_matrix: List[List[int]]):
        pass

    def get_players(self) -> List[PlayerTV]:
        pass

    def in_previous_rounds(self: SubsessionTV) -> List[SubsessionTV]:
        pass

    def in_all_rounds(self: SubsessionTV) -> List[SubsessionTV]:
        pass

    def in_round(self: SubsessionTV, round_number) -> SubsessionTV:
        pass

    def in_rounds(self: SubsessionTV, first, last) -> List[SubsessionTV]:
        pass

    def group_like_round(self, round_number: int):
        pass

    def group_randomly(self, fixed_id_in_group: bool = False):
        pass

    def field_maybe_none(self, field_name: str):
        pass

# Using TypeVar instead of the BaseSubsession seems to make PyCharm
# allow BaseSubsession be passed to a function marked as taking a Subsession arg
SubsessionTV = TypeVar("SubsessionTV", bound=BaseSubsession)

class BaseGroup:
    session: Any
    subsession: BaseSubsession
    round_number: int
    id_in_subsession: int
    def get_players(self) -> List[PlayerTV]:
        pass

    def get_player_by_role(self, role) -> PlayerTV:
        pass

    def get_player_by_id(self, id_in_group) -> PlayerTV:
        pass

    def in_previous_rounds(self: GroupTV) -> List[GroupTV]:
        pass

    def in_all_rounds(self: GroupTV) -> List[GroupTV]:
        pass

    def in_round(self: GroupTV, round_number) -> GroupTV:
        pass

    def in_rounds(self: GroupTV, first: int, last: int) -> List[GroupTV]:
        pass

    def field_maybe_none(self, field_name: str):
        pass

    def field_display(self, field_name: str):
        pass

GroupTV = TypeVar("GroupTV", bound=BaseGroup)

class BasePlayer:
    id_in_group: int
    # remove this from autocomplete because it crowds out id_in_group which is much more relevant
    # id_in_subsession: int
    payoff: Currency
    participant: Any
    session: Any
    group: GroupTV
    subsession: BaseSubsession
    round_number: int
    role: str
    def in_previous_rounds(self: PlayerTV) -> List[PlayerTV]:
        pass

    def in_all_rounds(self: PlayerTV) -> List[PlayerTV]:
        pass

    def get_others_in_group(self: PlayerTV) -> List[PlayerTV]:
        pass

    def get_others_in_subsession(self: PlayerTV) -> List[PlayerTV]:
        pass

    def in_round(self: PlayerTV, round_number) -> PlayerTV:
        pass

    def in_rounds(self: PlayerTV, first, last) -> List[PlayerTV]:
        pass

    def field_maybe_none(self, field_name: str):
        pass

    def field_display(self, field_name: str):
        pass

PlayerTV = TypeVar("PlayerTV", bound=BasePlayer)

T_extramodel = TypeVar('T_extramodel')

class ExtraModel:
    @classmethod
    def filter(cls: Type[T_extramodel], **kwargs) -> List[T_extramodel]:
        pass

    @classmethod
    def create(cls: Type[T_extramodel], **kwargs) -> T_extramodel:
        pass
    id: int

class WaitPage:
    wait_for_all_groups = False
    group_by_arrival_time = False
    title_text: str
    body_text: str
    template_name: str
    preserve_unsubmitted_inputs: bool

    round_number: int
    @staticmethod
    def is_displayed(player: Player):
        pass

    @staticmethod
    def js_vars(player: Player):
        pass

    @staticmethod
    def vars_for_template(player: Player):
        pass

    @staticmethod
    def app_after_this_page(player: Player, upcoming_apps):
        pass

    @staticmethod
    def after_all_players_arrive(group: Group):
        pass

    @staticmethod
    def live_method(player: Player, data):
        pass

class Page:
    round_number: int
    template_name: str
    timeout_seconds: int
    timer_text: str
    form_model: str
    form_fields: List[str]
    preserve_unsubmitted_inputs: bool
    allow_back_button: bool

    @staticmethod
    def live_method(player: Player, data):
        pass

    @staticmethod
    def get_form_fields(player: Player):
        pass

    @staticmethod
    def vars_for_template(player: Player):
        pass

    @staticmethod
    def js_vars(player: Player):
        pass

    @staticmethod
    def before_next_page(player: Player, timeout_happened):
        pass

    @staticmethod
    def is_displayed(player: Player):
        pass

    @staticmethod
    def error_message(player: Player, values):
        pass

    @staticmethod
    def get_timeout_seconds(player: Player):
        pass

    @staticmethod
    def app_after_this_page(player: Player, upcoming_apps):
        pass

class Bot:
    html: str
    case: Any
    cases: List
    participant: Any
    session: Any
    round_number: int
    player: PlayerTV
    group: GroupTV
    subsession: SubsessionTV

def Submission(
    PageClass, post_data: dict = {}, *, check_html=True, timeout_happened=False
):
    pass

def SubmissionMustFail(
    PageClass, post_data: dict = {}, *, check_html=True, error_fields=[]
):
    pass

def expect(*args):
    pass

# if i use Type[AnyModel] it doesn't get recognized as a superclass
# app_models = Union[Type[ExtraModel], Type[BasePlayer], Type[BaseGroup], Type[BaseSubsession]]
app_models = Union[
    Type[ExtraModel], Type[BasePlayer], Type[BaseGroup], Type[BaseSubsession]
]

def read_csv(path: str, type_model: app_models) -> List[dict]:
    pass

def render_template(path: str, *, player, C, vars: dict):
    pass

__all__ = [
    "Currency",
    "cu",
    "currency_range",
    "models",
    "widgets",
    "BaseConstants",
    "BaseSubsession",
    "BaseGroup",
    "BasePlayer",
    "ExtraModel",
    "WaitPage",
    "Page",
    "Bot",
    "Submission",
    "SubmissionMustFail",
    "expect",
    "read_csv",
    "render_template",
]



from argparse import ArgumentParser
from importlib import import_module
import sys


class BaseCommand:
    def outer_handle(self, args):
        parser = self._create_parser()
        options = parser.parse_args(args)
        return self.handle(**vars(options))

    def handle(self, *args, **options):
        raise NotImplementedError

    def _create_parser(self):
        parser = ArgumentParser()
        self.add_arguments(parser)
        return parser

    def add_arguments(self, parser):
        """
        Entry point for subclassed cli to add custom arguments.
        """
        pass


def call_command(cmd, *args):
    try:
        module = import_module(f'otree.cli.{cmd}')
    except ModuleNotFoundError:
        sys.exit(f"No command named '{cmd}'")
    module.Command().outer_handle(args)



from otree.common import signer_sign
import asyncio
from collections import defaultdict
from typing import DefaultDict, Dict
from urllib.parse import urlencode
import websockets.exceptions
from starlette.websockets import WebSocket

from otree.common import signer_sign
from otree.common2 import json_dumps


def wrap_websocket_send(original_send):
    async def send(message):
        try:
            await original_send(message)
        except websockets.exceptions.ConnectionClosed:
            # "You can catch and handle ConnectionClosed to prevent it from being logged."
            # https://websockets.readthedocs.io/en/latest/howto/faq.html#what-does-connectionclosederror-no-close-frame-received-or-sent-mean
            pass

    return send


class ChannelLayer:
    _subs: DefaultDict[str, Dict[int, WebSocket]]

    def _get_sockets(self, group):
        for socket in self._subs[group].values():
            yield socket

    def __init__(self):
        self._subs = defaultdict(dict)

    def add(self, group: str, websocket: WebSocket):
        self._subs[group][id(websocket)] = websocket

    def discard(self, group, websocket):
        group_dict = self._subs[group]
        group_dict.pop(id(websocket), None)
        # prune it so this global var doesn't grow indefinitely
        if not group_dict:
            del self._subs[group]

    async def send(self, group, data):
        data_json = json_dumps(data)
        tasks = [socket.send_text(data_json) for socket in self._get_sockets(group)]
        if tasks:
            await asyncio.gather(*tasks, return_exceptions=True)

    def sync_send(self, group, data):
        # Fire-and-forget approach to avoid blocking lock2.
        # By scheduling the task instead of waiting, we release the global lock faster,
        # allowing higher throughput. The send happens asynchronously in the event loop.
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            # Fallback for tests or non-async contexts
            asyncio.run(self.send(group, data))
        else:
            asyncio.create_task(self.send(group, data))


channel_layer = ChannelLayer()


async def group_send(*, group: str, data: dict):
    await channel_layer.send(group, data)


def sync_group_send(*, group: str, data: dict):
    channel_layer.sync_send(group=group, data=data)


def group_wait_page_name(session_id, page_index, group_id):

    return 'wait-page-{}-page{}-{}'.format(session_id, page_index, group_id)


def subsession_wait_page_name(session_id, page_index):

    return 'wait-page-{}-page{}'.format(session_id, page_index)


def gbat_group_name(session_id, page_index):
    return 'group_by_arrival_time_session{}_page{}'.format(session_id, page_index)


def gbat_path(**kwargs):
    return '/group_by_arrival_time?' + urlencode(kwargs)


def room_participants_group_name(room_name):
    return 'room-participants-{}'.format(room_name)


def room_participant_path(**kwargs):
    return '/wait_for_session_in_room?' + urlencode(kwargs)


def session_monitor_group_name(session_code):
    return f'session-monitor-{session_code}'


def session_monitor_path(session_code):
    return f'/session_monitor/{session_code}'


def room_admin_group_name(room_name):
    return f'room-admin-{room_name}'


def room_admin_path(room_name):
    return '/room_without_session/{}'.format(room_name)


def create_session_path():
    return '/create_session'


def create_demo_session_path():
    return '/create_demo_session'


def group_wait_page_path(**kwargs):
    return '/wait_page?' + urlencode(kwargs)


def subsession_wait_page_path(**kwargs):
    return '/subsession_wait_page?' + urlencode(kwargs)


def browser_bots_launcher_group(session_code):
    return 'browser-bots-client-{}'.format(session_code)


def browser_bots_launcher_path(session_code):
    return '/browser_bots_client/{}'.format(session_code)


def auto_advance_path(**kwargs):
    return '/auto_advance?' + urlencode(kwargs)


def auto_advance_group(participant_code):
    return f'auto-advance-{participant_code}'


def live_group(session_code, page_index, pcode):
    '''
    live_method_hash is so that you can send messages across pages that share the same
    live_method. But you don't want to send messages to a different live_method page.
    '''
    return f'live-{session_code}-{page_index}-{pcode}'


def live_path(**kwargs):
    return f'/live?' + urlencode(kwargs)


def trial_group(pcode, page_index):
    return f'trial-{pcode}-{page_index}'


def trial_path(**kwargs):
    return f'/trial?' + urlencode(kwargs)


def chat_path(channel, participant_id):
    return '/chat?' + urlencode(
        {
            'channel': signer_sign(channel),
            'participant_id': signer_sign(str(participant_id)),
        }
    )


def get_chat_group(channel):
    return 'otreechat-{}'.format(channel)



import asyncio
import time
from typing import Dict, Any
import asyncio
import otree.common
from otree.channels import utils as channel_utils
from otree.models import Participant, BasePlayer, BaseGroup
from otree.lookup import get_page_lookup
import logging
from otree.database import NoResultFound
import inspect

logger = logging.getLogger(__name__)


def _get_live_page_entities(participant_code):
    try:
        participant = Participant.objects_get(code=participant_code)
    except NoResultFound:
        logger.warning(f'Participant not found: {participant_code}')
        return
    lookup = get_page_lookup(participant._session_code, participant._index_in_pages)
    app_name = lookup.app_name
    models_module = otree.common.get_models_module(app_name)
    PageClass = lookup.page_class
    return app_name, lookup, models_module, PageClass, participant


def live_method_is_async(participant_code):
    app_name, lookup, models_module, PageClass, participant = _get_live_page_entities(
        participant_code
    )
    return inspect.isasyncgenfunction(PageClass.live_method)


async def live_payload_function(participant_code, page_name, payload):

    app_name, lookup, models_module, PageClass, participant = _get_live_page_entities(
        participant_code
    )
    # this could be incorrect if the player advances right after liveSend is executed.
    # maybe just return if it doesn't match. (but leave it in for now and see how much that occurs,
    # don't want silent failures.)
    if page_name != PageClass.__name__:
        logger.warning(
            f'Ignoring liveSend message from {participant_code} because '
            f'they are on page {PageClass.__name__}, not {page_name}.'
        )
        return

    player = models_module.Player.objects_get(
        round_number=lookup.round_number, participant=participant
    )

    # it makes sense to check the group first because
    # if the player forgot to define it on the Player,
    # we shouldn't fall back to checking the group. you could get an error like
    # 'Group' has no attribute 'live_auction' which would be confusing.
    # also, we need this 'group' object anyway.
    # and this is a good place to show the deprecation warning.
    group = player.group
    live_method = PageClass.live_method

    Player: BasePlayer = models_module.Player
    pcodes_dict = {
        d[0]: d[1]
        for d in Player.objects_filter(group=group)
        .join(Participant)
        .with_entities(
            Player.id_in_group,
            Participant.code,
        )
    }

    async for retval in call_live_method_compat(live_method, player, payload):

        # we should require a return, otherwise the user might forget it.
        # but this breaks backward compat.
        if not retval:
            return

        if not isinstance(retval, dict):
            msg = f'live method must return a dict'
            raise LiveMethodBadReturnValue(msg)

        if 0 in retval:
            if len(retval) > 1:
                raise LiveMethodBadReturnValue(
                    'If dict returned by live_method has key 0, it must not contain any other keys'
                )
        else:
            for pid in retval:
                if pid not in pcodes_dict:
                    msg = f'live_method has invalid return value. No player with id_in_group={repr(pid)}'
                    raise LiveMethodBadReturnValue(msg)

        pcode_retval = {}
        for pid, pcode in pcodes_dict.items():
            payload = retval.get(pid, retval.get(0))
            if payload is not None:
                pcode_retval[pcode] = {
                    'otree_success': True,
                    'live_method_payload': payload,
                }

        await _live_send_back(
            participant._session_code, participant._index_in_pages, pcode_retval
        )


class LiveMethodBadReturnValue(Exception):
    pass


async def _live_send_back(session_code, page_index, pcode_retval):
    '''separate function for easier patching'''

    tasks = [
        channel_utils.group_send(
            group=channel_utils.live_group(session_code, page_index, pcode),
            data=retval,
        )
        for pcode, retval in pcode_retval.items()
    ]
    if tasks:
        await asyncio.gather(*tasks, return_exceptions=True)


ASYNC_OR_YIELD_ISSUE_MSG = "live_method can only be (a) a regular function that returns a value, or (b) an async generator function."


async def call_live_method_compat(live_method, player, payload):
    """
    before the "compat" referred to string vs method.
    now it refers to yield/return style
    """

    # locally the server is not shutting down properly when i do Ctrl+C,
    # but that was a pre-existing issue with my local dev setup, apparently.
    if inspect.isasyncgenfunction(live_method):
        # Case 1: async generator function (async def with yield)
        try:
            async for item in live_method(player, payload):
                yield item
        except asyncio.CancelledError:
            # Handle graceful cancellation when server shuts down
            # this is needed to allow Ctrl+C to stop the current tasks.
            return
    elif inspect.iscoroutinefunction(live_method):
        # Async function (async def without yield) - not allowed
        raise LiveMethodBadReturnValue(ASYNC_OR_YIELD_ISSUE_MSG)
    else:
        # Case 2: regular function
        result = live_method(player, payload)
        if inspect.isgenerator(result):
            raise LiveMethodBadReturnValue(ASYNC_OR_YIELD_ISSUE_MSG)
        yield result


_USER_LOCK_MAX_IDLE_TIME = 3600


class ParticipantLockManager:
    def __init__(self):
        self._locks: Dict[str, asyncio.Lock] = {}
        self._last_used: Dict[str, float] = {}

    def get_lock(self, pcode: str) -> asyncio.Lock:
        now = time.time()

        # Clean on every request (very lightweight)
        self._cleanup_expired(now)

        # Get or create lock
        if pcode not in self._locks:
            self._locks[pcode] = asyncio.Lock()

        self._last_used[pcode] = now
        return self._locks[pcode]

    def _cleanup_expired(self, now: float):
        expired = [
            pcode
            for pcode, last_used in self._last_used.items()
            if (
                now - last_used > _USER_LOCK_MAX_IDLE_TIME
                and not self._locks[pcode].locked()
            )
        ]

        for pcode in expired:
            del self._locks[pcode]
            del self._last_used[pcode]


_lock_manager = ParticipantLockManager()


def get_participant_scoped_lock(pcode: str) -> asyncio.Lock:
    return _lock_manager.get_lock(pcode)



from otree.database import st, CurrencyType, Currency


def read_csv_bool(val):
    if val == '':
        return None
    return val in ['TRUE', '1', 'True', 'true', 1]


def read_csv_int(val):
    if val == '':
        return None
    return int(val)


def read_csv_float(val):
    if val == '':
        return None
    return float(val)


def read_csv_currency(val):
    if val == '':
        return None
    return Currency(val)


def read_csv_str(val):
    # should '' be interpreted as empty string or None?
    # i think empty string is better.
    # (1) the principle that you should not have 2 values for None,
    # (2) because this avoids null reference errors. you can use all string operations on an empty string.
    # it's true that oTree models use None as the default value for a StringField,
    # but that seems a bit different to me.
    return str(val)


def map_types(d, mapping: dict):
    ret = {}
    for k, v in d.items():
        type_conversion_function = mapping[k]
        try:
            ret[k] = type_conversion_function(v)
        except Exception:
            msg = f"CSV file contains an incompatible value in column '{k}': {repr(v)}"
            raise Exception(msg) from None
    return ret


class MissingFieldError(Exception):
    pass


def read_csv(path: str, type_model):
    import csv

    CONVERSION_FUNCTIONS = {
        st.Boolean: read_csv_bool,
        CurrencyType: read_csv_currency,
        st.Float: read_csv_float,
        st.Integer: read_csv_int,
        st.String: read_csv_str,
        st.Text: read_csv_str,
    }

    # even if it's not going into an ExtraModel, you can still make a model just for the purpose of CSV loading.
    with open(path, 'r', encoding='utf-8-sig') as f:
        # detect whether it's delimited with "," or ";"
        # (Excel defaults to ";" if the user's locale is German etc.)
        dialect = csv.Sniffer().sniff(f.read(1024), delimiters=";,")
        dialect.strict = True
        f.seek(0)
        reader = csv.DictReader(f, dialect=dialect)
        mapping = {}
        for fieldname in reader.fieldnames:
            try:
                _coltype = type(type_model.__table__.columns[fieldname].type)
            except KeyError as exc:
                # it's good to be strict and require all columns. This will prevent issues like
                # typos and users simply not understanding how the feature works.
                model_name = type_model.__name__
                msg = f"CSV file contains column '{exc.args[0]}', which is not found in model {model_name}."
                raise MissingFieldError(msg) from None
            mapping[fieldname] = CONVERSION_FUNCTIONS[_coltype]

        return [map_types(row, mapping) for row in reader]



from starlette.routing import WebSocketRoute as WSR

from . import consumers as cs

# careful about this import. it seems to mess with the db initialization that makes my tests fail.
# import otree.trials

websocket_routes = [
    WSR('/wait_page', cs.WSGroupWaitPage),
    WSR('/subsession_wait_page', cs.WSSubsessionWaitPage),
    WSR('/group_by_arrival_time', cs.WSGroupByArrivalTime),
    WSR('/auto_advance', cs.DetectAutoAdvance),
    WSR('/create_session', cs.WSCreateSession),
    WSR('/create_demo_session', cs.WSCreateDemoSession),
    WSR('/delete_sessions', cs.WSDeleteSessions),
    WSR('/wait_for_session_in_room', cs.WSRoomParticipant),
    WSR('/room_without_session/{room_name}', cs.WSRoomAdmin),
    WSR('/session_monitor/{code}', cs.WSSessionMonitor),
    WSR('/browser_bots_client/{session_code}', cs.WSBrowserBotsLauncher),
    WSR('/browser_bot_wait', cs.WSBrowserBot),
    WSR('/live', cs.LiveConsumer),
    # WSR('/trial', otree.trials.TrialConsumer),
    WSR('/chat', cs.WSChat),
    WSR('/export', cs.WSExportData),
]



from io import StringIO
import datetime
from starlette.responses import Response
from . import cbv
from starlette.endpoints import HTTPEndpoint

import otree.common
import otree.models
from otree import export
from otree.export import BOM, get_installed_apps_with_data
from otree.models import Participant
from otree.models_concrete import ChatMessage
from otree.database import dbq
from . import cbv


def get_csv_http_response(buffer: StringIO, filename_prefix) -> Response:
    buffer.seek(0)
    response = Response(buffer.read())
    date = datetime.date.today().isoformat()
    response.headers['Content-Type'] = 'text/csv'
    response.headers['Content-Disposition'] = (
        f'attachment; filename="{filename_prefix}-{date}.csv"'
    )
    return response


class ExportMixin:
    """Shared logic for all export views"""

    def file_name_stem(self):
        raise NotImplementedError

    def write_to_buf(self, buf, session_code, query_params):
        raise NotImplementedError

    def process_export(self, request):
        query_params = request.query_params
        session_code = query_params.get('session_code')
        buf = StringIO()
        if query_params.get('format') == 'csv_bom':
            buf.write(BOM)
        self.write_to_buf(buf, session_code, query_params)
        return get_csv_http_response(buf, self.file_name_stem())


class BaseExport(ExportMixin, HTTPEndpoint):
    """REST API export with REST key authentication"""

    def get(self, request):
        error_response = cbv.validate_rest_key(request)
        if error_response:
            return error_response
        return self.process_export(request)


class LoginProtectedExport(ExportMixin, cbv.AdminView):
    """Login-protected export for non-API endpoints"""

    def get(self, request, **kwargs):
        return self.process_export(request)


class ExportWide(BaseExport):
    '''used by data page'''

    url_pattern = '/api/export_wide'

    def file_name_stem(self):
        return 'all_apps_wide'

    def write_to_buf(self, buf, session_code, query_params):
        export.export_wide(buf, session_code=session_code)


class ExportPageTimes(BaseExport):

    url_pattern = '/api/export_page_times'

    def file_name_stem(self):
        return 'PageTimes'

    def write_to_buf(self, buf, session_code, query_params):
        export.export_page_times(buf)


class ExportChat(BaseExport):

    url_pattern = '/api/export_chat'

    def file_name_stem(self):
        return 'ChatMessages'

    def write_to_buf(self, buf, session_code, query_params):
        export.export_chat(buf, session_code=session_code)


class ExportApp(BaseExport):

    url_pattern = '/api/export_app'

    def file_name_stem(self):
        return self.app_name

    def write_to_buf(self, buf, session_code, query_params):
        app_name = query_params.get('app')
        self.app_name = app_name
        if not app_name:
            raise ValueError("app parameter is required")
        export.export_app(app_name, buf, session_code=session_code)


class ExportAppCustom(BaseExport):

    url_pattern = '/api/export_app_custom'

    app_name: str
    function_name: str

    def file_name_stem(self):
        return f'{self.app_name}_{self.function_name}'

    def write_to_buf(self, buf, session_code, query_params):
        app_name = query_params.get('app')
        function_name = query_params.get('function_name', 'custom_export')
        self.app_name = app_name
        self.function_name = function_name
        if not app_name:
            raise ValueError("app parameter is required")
        export.custom_export_app(
            app_name, buf, session_code=session_code, function_name=function_name
        )


class ExportPageTimesLogin(LoginProtectedExport):
    """Login-protected version of page times export"""

    url_pattern = '/ExportPageTimes'

    def file_name_stem(self):
        return 'PageTimes'

    def write_to_buf(self, buf, session_code, query_params):
        export.export_page_times(buf)


class ExportChatLogin(LoginProtectedExport):
    """Login-protected version of chat export"""

    url_pattern = '/ExportChat'

    def file_name_stem(self):
        return 'ChatMessages'

    def write_to_buf(self, buf, session_code, query_params):
        export.export_chat(buf, session_code=session_code)


class ExportIndex(cbv.AdminView):
    url_pattern = '/ExportIndex'

    def vars_for_template(self):
        from otree.asgi import reverse

        # can't use settings.OTREE_APPS, because maybe the app
        # was removed from SESSION_CONFIGS.
        app_names_with_data = get_installed_apps_with_data()

        # Find all custom export functions
        custom_export_list = []
        for app_name in app_names_with_data:
            custom_exports = export.get_custom_export_functions(app_name)
            for export_function_name in sorted(custom_exports.keys()):
                custom_export_list.append((app_name, export_function_name))

        # Base URLs
        wide_url = reverse('ExportWide')
        export_url = reverse('ExportApp')
        custom_export_url = reverse('ExportAppCustom')
        page_times_url = reverse('ExportPageTimes')
        chat_url = reverse('ExportChat')

        # Wide export table
        wide_table = [
            dict(
                name='All apps (wide format)',
                csv_url=wide_url,
                excel_url=wide_url + '?format=csv_bom',
                example_session_url=wide_url + '?session_code={SESSION_CODE}',
            )
        ]

        # Per-app export table
        app_table = [
            dict(
                name=app_name,
                csv_url=export_url + '?app=' + app_name,
                excel_url=export_url + '?format=csv_bom&app=' + app_name,
                example_session_url=export_url
                + '?app='
                + app_name
                + '&session_code={SESSION_CODE}',
                app_name=app_name,
            )
            for app_name in app_names_with_data
        ]

        # Custom export table
        custom_export_table = []
        for app_name, export_function_name in custom_export_list:
            custom_export_table.append(
                dict(
                    name=f'{app_name} ({export_function_name})',
                    csv_url=custom_export_url
                    + f'?app={app_name}&function_name={export_function_name}',
                    excel_url=custom_export_url
                    + f'?format=csv_bom&app={app_name}&function_name={export_function_name}',
                    example_session_url=custom_export_url
                    + f'?app={app_name}&function_name={export_function_name}'
                    + '&session_code={SESSION_CODE}',
                    app_name=app_name,
                    function_name=export_function_name,
                )
            )

        # Other exports (using login-protected endpoints)
        page_times_login_url = reverse('ExportPageTimesLogin')
        chat_login_url = reverse('ExportChatLogin')

        other_exports = [
            dict(
                name='Page times',
                csv_url=page_times_login_url,
                excel_url=page_times_login_url + '?format=csv_bom',
                api_csv_url=page_times_url,
                api_excel_url=page_times_url + '?format=csv_bom',
                example_session_url=None,
            ),
        ]

        chat_messages_exist = bool(dbq(ChatMessage).first())
        if chat_messages_exist:
            other_exports.append(
                dict(
                    name='Chat logs',
                    csv_url=chat_login_url,
                    excel_url=chat_login_url + '?format=csv_bom',
                    api_csv_url=chat_url,
                    api_excel_url=chat_url + '?format=csv_bom',
                    example_session_url=chat_url + '?session_code={SESSION_CODE}',
                )
            )

        return dict(
            db_is_empty=not bool(dbq(Participant).first()),
            wide_table=wide_table,
            app_table=app_table,
            custom_export_table=custom_export_table,
            other_exports=other_exports,
        )



import gettext as gettext_lib
from otree.common import FULL_DECIMAL_PLACES
from otree import settings
import re


# these symbols are a fallback if we don't have an explicit rule
# for the currency/language combination.
# (most common situation is where the language is English)

CURRENCY_SYMBOLS = {
    'AED': 'AED',
    'ARS': '$',
    'AUD': '$',
    'BRL': 'R$',
    'CAD': '$',
    'CHF': 'CHF',
    # need to use yuan character here, that's what gets shown
    # on form inputs. but if you run a study in english, it will
    # still show 元, which is not ideal. but that is rare.
    'CNY': '元',
    'CZK': 'Kč',
    'DKK': 'kr',
    'EGP': 'ج.م.‏',
    'EUR': '€',
    'GBP': '£',
    'HKD': 'HK$',
    'HUF': 'Ft',
    'ILS': '₪',
    'INR': '₹',
    'JPY': '円',
    'KRW': '원',
    'MXN': '$',
    'MYR': 'RM',
    'NOK': 'kr',
    'PLN': 'zł',
    'RUB': '₽',
    'SEK': 'kr',
    'SGD': 'SGD',
    'THB': 'THB',
    'TRY': '₺',
    'TWD': '$',
    'USD': '$',
    'ZAR': 'R',
}


def get_currency_format(lc: str, LO: str, CUR: str) -> str:
    '''because of all the if statements, this has very low code coverage
    but it's ok'''

    ##############################
    # Languages with complex rules
    ##############################

    if lc == 'en':
        if CUR in ['USD', 'CAD', 'AUD']:
            return '$#'
        if CUR == 'GBP':
            return '£#'
        if CUR == 'EUR':
            return '€#'
        if CUR == 'INR':
            return '₹ #'
        if CUR == 'SGD':
            return '$#'
        # override for CNY/JPY/KRW, otherwise it would be written as 원10
        # need to use the chinese character because that's already what's used in
        # form inputs
        if CUR == 'CNY':
            return '#元'
        if CUR == 'JPY':
            return '#円'
        if CUR == 'KRW':
            return '#원'
        if CUR == 'ZAR':
            return 'R#'
        return '¤#'

    if lc == 'zh':
        if CUR == 'CNY':
            return '#元'
        if CUR == 'HKD':
            return 'HK$#'
        if CUR == 'TWD':
            return '$#'
        if CUR == 'SGD':
            return 'SGD#'
        return '¤#'

    if lc == 'de':
        if CUR == 'EUR':
            if LO == 'AT':
                return '€ #'
            return '# €'
        if CUR == 'CHF':
            return 'CHF #'
        return '¤ #'

    if lc == 'es':
        if CUR == 'ARS':
            return '$ #'
        if CUR == 'EUR':
            return '# €'
        if CUR == 'MXN':
            return '$#'
        return '# ¤'

    if lc == 'nl':
        if LO == 'BE':
            if CUR == 'EUR':
                return '# €'
            return '# ¤'
        # default NL
        if CUR == 'EUR':
            return '€ #'
        return '¤ #'

    if lc == 'pt':
        if CUR == 'BRL':
            return 'R$#'
        if CUR == 'EUR':
            return '# €'
        return '¤#'

    if lc == 'ar':
        if CUR == 'AED':
            return 'د.إ.‏ #'
        return '¤ #'

    #############################
    # Languages with simple rules
    #############################

    if lc == 'cs':
        if CUR == 'CZK':
            return '# Kč'
        return '# ¤'
    if lc == 'da':
        if CUR == 'DKK':
            return '# kr.'
        return '# ¤'
    if lc == 'fi':
        if CUR == 'EUR':
            return '# €'
        return '# ¤'
    if lc == 'fr':
        if CUR == 'EUR':
            return '# €'
        return '# ¤'
    if lc == 'he':
        if CUR == 'ILS':
            return '# ₪'
        return '# ¤'
    if lc == 'hu':
        if CUR == 'HUF':
            return '# Ft'
        return '# ¤'
    if lc == 'it':
        if CUR == 'EUR':
            return '# €'
        return '# ¤'
    if lc == 'ja':
        if CUR == 'JPY':
            return '#円'
        return '¤#'
    if lc == 'ko':
        if CUR == 'KRW':
            return '#원'
        return '¤#'
    if lc == 'ms':
        if CUR == 'MYR':
            return 'RM#'
        return '¤#'
    if lc == 'nb':
        if CUR == 'NOK':
            return 'kr #'
        return '¤ #'
    if lc == 'pl':
        if CUR == 'PLN':
            return '# zł'
        return '# ¤'
    if lc == 'ru':
        if CUR == 'RUB':
            return '# ₽'
        return '# ¤'
    if lc == 'sv':
        if CUR == 'SEK':
            return '# kr'
        return '# ¤'
    if lc == 'th':
        if CUR == 'THB':
            return '฿#'
        return '¤#'
    if lc == 'tr':
        if CUR == 'TRY':
            return '# ₺'
        return '# ¤'
    if lc == 'hi':
        if CUR == 'INR':
            return '₺#'
        return '¤#'

    # fallback if it's another language, etc.
    return '# ¤'


# def format_number(number, *, places):
#     """we don't use locale.setlocale because e.g.
#     only english locale is installed on heroku
#
#     This is a complex function because it's is used by many different things.
#     - formatting any number (random floats, etc)
#     - forms
#     - to0, to1, to2
#
#     """
#     # we use FULL_DECIMAL_PLACES as the arg because it's more explict than None.
#     if places is FULL_DECIMAL_PLACES:
#         places = None
#     str_number = str(number)
#     if '.' in str_number:
#         lhs, rhs = str_number.split('.')
#     else:
#         lhs = str_number
#         if places is None:
#             return lhs
#         rhs = ''
#     if places is not None:
#         rhs = rhs.ljust(places, '0')
#     if places == 0:
#         return lhs
#     # rhs[:None] just takes the whole thing, which is the desired behavior
#     # with floats.
#     return lhs + settings.DECIMAL_SEPARATOR + rhs[:places]


def _convert_decimal_separator_char(c):
    # need to do it in 1 pass
    if c == ',':
        return settings.THOUSAND_SEPARATOR
    if c == '.':
        return settings.DECIMAL_SEPARATOR
    return c


def convert_decimal_separator(num_str):
    return ''.join(_convert_decimal_separator_char(c) for c in num_str)


def format_number(number, *, places):
    if places == FULL_DECIMAL_PLACES:
        fmt = "{:,}"
    else:
        fmt = "{:,.%df}" % places
    res = fmt.format(number)
    return convert_decimal_separator(res)


def format_number_for_form(number, *, places):
    if places == FULL_DECIMAL_PLACES:
        fmt = "{}"
    else:
        fmt = "{:.%df}" % places
    res = fmt.format(number)
    return convert_decimal_separator(res)


def extract_otreetemplate(fileobj, keywords, comment_tags, options):
    """Deprecated"""
    for lineno, line in enumerate(fileobj, start=1):
        for msg in re.findall(r"""\{\{\s?trans ['"](.*)['"]\s?\}\}""", line.decode()):
            yield (lineno, 'trans', msg, [])


def extract_otreetemplate_internal(fileobj, keywords, comment_tags, options):
    """babel custom extractor for |gettext in otree templates"""
    for lineno, line in enumerate(fileobj, start=1):
        for msg in re.findall(
            r"""\{\{\s?['"](.*)['"]\|gettext\s?\}\}""", line.decode()
        ):
            # not sure exactly what the 2nd arg is for
            yield (lineno, 'gettext', msg, [])


def core_gettext(msg):
    return gettext_lib.dgettext('django', msg)


"""
Don't use core-ngettext because pybabel doesn't recognize that it is for
plurals 
"""



import inspect
import os
from importlib import import_module
from pathlib import Path
import sys
from otree import common
from otree.api import (
    BasePlayer,
    BaseGroup,
    BaseSubsession,
    Currency,
    WaitPage,
)
from otree.views.abstract import Page
from otree import settings
from otree.common import get_pages_module, get_models_module, get_builtin_constant
from collections import namedtuple


Error = namedtuple(
    'Error',
    [
        'app_name',
        'title',
        'id',
    ],
)
Warning = namedtuple(
    'Warning',
    [
        'app_name',
        'title',
        'id',
    ],
)

print_function = print


class AppCheckHelper:
    def __init__(self, app_name):
        self.app_name = app_name
        self.path = Path(app_name)
        self.errors = []
        self.warnings = []

    def add_error(self, title, numeric_id: int):
        self.errors.append(Error(title=title, id=numeric_id, app_name=self.app_name))

    def add_warning(self, title, numeric_id: int):
        self.warnings.append(
            Warning(title=title, id=numeric_id, app_name=self.app_name)
        )

    def get_template_names(self):
        templates_dir = self.path / 'templates'
        for root, dirs, files in os.walk(templates_dir):
            for name in files:
                if name.endswith('.html'):
                    yield templates_dir.joinpath(root, name)

    def get_path(self, name):
        return self.path.joinpath(name)

    def module_exists(self, module):
        return self.path.joinpath(module + '.py').exists()


base_model_attrs = {
    'Player': set(dir(BasePlayer)),
    'Group': set(dir(BaseGroup)),
    'Subsession': set(dir(BaseSubsession)),
}
model_field_substitutes = {
    int: 'IntegerField',
    float: 'FloatField',
    bool: 'BooleanField',
    str: 'CharField',
    Currency: 'CurrencyField',
    type(None): 'IntegerField',
    # not always int, but it's a reasonable suggestion
}


def model_classes(helper: AppCheckHelper, app_name):
    models = get_models_module(app_name)
    for name in ['Subsession', 'Group', 'Player']:
        if not hasattr(models, name):
            helper.add_error(
                'MissingModel: Model "%s" not defined' % name, numeric_id=110
            )

    Player = models.Player
    Group = models.Group
    Subsession = models.Subsession

    if hasattr(Subsession, 'before_session_starts'):
        msg = (
            'before_session_starts no longer exists. '
            "You should rename it to creating_session."
        )
        helper.add_error(msg, numeric_id=119)

    for Model in [Player, Group, Subsession]:
        for attr_name in dir(Model):
            if attr_name not in base_model_attrs[Model.__name__]:
                attr_value = getattr(Model, attr_name)
                _type = type(attr_value)
                if _type in model_field_substitutes.keys():
                    msg = (
                        'NonModelFieldAttr: '
                        '{model} has attribute "{attr}", which is not a model field, '
                        'and will therefore not be saved '
                        'to the database. '
                        'Consider changing to "{attr} = models.{FieldType}(initial={attr_value})"'
                    ).format(
                        model=Model.__name__,
                        attr=attr_name,
                        FieldType=model_field_substitutes[_type],
                        attr_value=repr(attr_value),
                    )
                    helper.add_error(msg, numeric_id=111)

                # if people just need an iterable of choices for a model field,
                # they should use a tuple, not list or dict
                elif _type in {list, dict, set}:
                    warning = (
                        'MutableModelClassAttr: '
                        '{ModelName}.{attr} is a {type_name}. '
                        'Modifying it during a session (e.g. appending or setting values) '
                        'will have unpredictable results; '
                        'you should use '
                        'session.vars or participant.vars instead. '
                        'Or, if this {type_name} is read-only, '
                        "then it's recommended to move it outside of this class "
                        '(e.g. put it in Constants).'
                    ).format(
                        ModelName=Model.__name__,
                        attr=attr_name,
                        type_name=_type.__name__,
                    )

                    helper.add_error(warning, numeric_id=112)


def is_builtin(func_name):
    if func_name in [
        'creating_session',
        'group_by_arrival_time_method',
        'vars_for_admin_report',
    ]:
        return True
    endings = ['_min', '_max', '_choices', '_error_message']
    for ending in endings:
        if func_name.endswith(ending):
            return True
    for beginning in ['custom_export']:
        if func_name.startswith(beginning):
            return True
    return False


def uncalled_functions(helper: AppCheckHelper, app_name):
    import re

    if not common.is_noself(app_name):
        return

    txt = Path(f'{app_name}/__init__.py').read_text('utf8')
    for match in re.finditer(r'def (\w+)\(', txt):
        func_name = match.group(1)
        if is_builtin(func_name):
            continue
        # hasattr( is a double-check in case the function is commented out
        # it's also ok if the function is used as as method, e.g. oTree Studio.
        # also, some functions are assigned as variables, e.g. is_displayed = is_displayed
        if txt.count(func_name) == 1:
            app = import_module(app_name)
            if hasattr(app, func_name):
                msg = f"'{func_name}' is defined but not used"
                helper.add_warning(msg, numeric_id=131)


def constants(helper: AppCheckHelper, app_name):

    models = get_models_module(app_name)

    if not hasattr(models, 'Constants') and not hasattr(models, 'C'):
        helper.add_error('App is missing a constants class', numeric_id=11)
        return

    attrs = ['name_in_url', 'players_per_group', 'num_rounds']
    for attr_name in attrs:
        try:
            get_builtin_constant(app_name, attr_name)
        except AttributeError:
            msg = "Constants class needs to define '{}'".format(attr_name)
            helper.add_error(msg.format(attr_name), numeric_id=12)
    ppg = get_builtin_constant(app_name, 'players_per_group')
    if ppg == 0 or ppg == 1:
        helper.add_error(
            "players_per_group cannot be {}. You "
            "should set it to None, which makes the group "
            "all players in the subsession.".format(ppg),
            numeric_id=13,
        )
    if ' ' in get_builtin_constant(app_name, 'name_in_url'):
        helper.add_error("name_in_url must not contain spaces", numeric_id=14)


def pages_function(helper: AppCheckHelper, app_name):
    pages_module = common.get_pages_module(app_name)
    try:
        page_list = pages_module.page_sequence
    except:
        helper.add_error('The variable page_sequence is missing.', numeric_id=21)
        return
    else:
        for i, ViewCls in enumerate(page_list):
            # there is no good reason to include Page in page_sequence.
            # As for WaitPage: even though it works fine currently
            # and can save the effort of subclassing,
            # we should restrict it, because:
            # - one user had "class WaitPage(Page):".
            # - if someone makes "class WaitPage(WaitPage):", they might
            #   not realize why it's inheriting the extra behavior.
            # overall, I think the small inconvenience of having to subclass
            # once per app
            # is outweighed by the unexpected behavior if someone subclasses
            # it without understanding inheritance.
            # BUT: built-in Trust game had a wait page called WaitPage.
            # that was fixed on Aug 24, 2017, need to wait a while...
            # see below in ensure_no_misspelled_attributes,
            # we can get rid of a check there also
            if ViewCls.__name__ == 'Page':
                msg = "page_sequence cannot contain a class called 'Page'."
                helper.add_error(msg, numeric_id=22)
            if ViewCls.__name__ == 'WaitPage' and app_name != 'trust':
                msg = "page_sequence cannot contain a class called 'WaitPage'."
                helper.add_error(msg, numeric_id=221)
            if issubclass(ViewCls, WaitPage):
                if ViewCls.group_by_arrival_time:
                    if i > 0:
                        helper.add_error(
                            '"{}" has group_by_arrival_time=True, so '
                            'it must be placed first in page_sequence.'.format(
                                ViewCls.__name__
                            ),
                            numeric_id=23,
                        )
                    if ViewCls.wait_for_all_groups:
                        helper.add_error(
                            'Page "{}" has group_by_arrival_time=True, so '
                            'it cannot have wait_for_all_groups=True also.'.format(
                                ViewCls.__name__
                            ),
                            numeric_id=24,
                        )
                    if hasattr(ViewCls, 'get_players_for_group'):
                        helper.add_error(
                            'Page "{}" defines get_players_for_group, which is deprecated. '
                            'You should instead define a top-level function called group_by_arrival_time_method. '
                            ''.format(ViewCls.__name__),
                            numeric_id=25,
                        )
            elif issubclass(ViewCls, Page):
                if ViewCls.has_trial():
                    Trial = ViewCls.trial_model
                    for fn in ViewCls.trial_stimulus_fields + getattr(
                        ViewCls, 'trial_response_fields', []
                    ):
                        if not hasattr(Trial, fn):
                            helper.add_error(
                                f"Page '{ViewCls.__name__}' includes '{fn}' in its trial fields, "
                                f"but this is not a field on the '{Trial.__name__}' model.",
                                numeric_id=27,
                            )
                    # i tried checking 'if Trial.player is not BaseTrial.player' here,
                    # but it's always unequal.
            else:
                msg = '"{}" is not a valid page'.format(ViewCls)
                helper.add_error(msg, numeric_id=26)


def get_checks_output(app_names=None):
    app_names = app_names or settings.OTREE_APPS
    errors = []
    warnings = []
    for check_function in [
        model_classes,
        constants,
        pages_function,
        uncalled_functions,
    ]:
        for app_name in app_names:
            helper = AppCheckHelper(app_name)
            check_function(helper, app_name)
            errors.extend(helper.errors)
            warnings.extend(helper.warnings)
    return errors, warnings


def run_checks():
    errors, warnings = get_checks_output()
    if errors:
        for ele in errors:
            print_function(ele)
        sys.exit(-1)
    for ele in warnings:
        print_function(ele)
